#!/usr/bin/env python

# Prerequisite: This program requires the maude binary 
# 
# This prototype tool was written by Sasa Radomirovic.  It converts a
# protocol given in Alice & Bob notation to role scripts that include
# explicit checks the protocol roles should perform on received
# messages when executing the protocol.
# The program code is an implementation of the algorithms described in 
# the paper "Alice and Bob Meet Equational Theories"
# by David Basin, Michel Keller, Sasa Radomirovic, and Ralf Sasse
# published in Logic, Rewriting, and Concurrency - Essays dedicated to Jose Meseguer on the Occasion of His 65th Birthday, pages 160-180, LNCS 9200, 2015.

# Example Input: 
# functions: pk/1, aenc/2, pair/2, adec/2, fst/1, snd/1
# equations: adec(aenc(M,pk(K)),K) = M
# 	     fst(pair(x,y)) = x
#	     snd(pair(x,y)) = y
#
# A knows A, skA, pk(skB)
# B knows B, skB, pk(skA)
# A: fresh N1 
# A -> B: aenc(pair(N1,A),pk(skB))
# B: fresh N2
# B -> A: aenc(pair(N1,N2),pk(skA))
# A -> B: aenc(N2,pk(skB))


from __future__ import print_function
import sys, os, subprocess
from commands import getoutput
from commands import getstatusoutput


maude_path = ""

### helper functions ###

def usage():
	str = """Usage:

{} file

This is a python script. It reads an Alice & Bob protocol 
specification from file or stdin and converts the specification 
into role scripts with explicit checks.
"""
	print(str.format(sys.argv[0]))

class bcolors:
    BLACK = '\033[90m'
    RED = '\033[91m'
    BOLDRED = '\033[1;31m'
    YELLOW = '\033[93m'
    YELLOWRED = '\033[0;33;41m'
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = YELLOWRED
    FAIL = BOLDRED
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    ENDC = '\033[0m'

# write an error on the stderr and quit the program 
def error(text):
	sys.stdout.flush()
	sys.stderr.write(bcolors.FAIL+"Error:"+bcolors.ENDC+" "+bcolors.BOLD+text+bcolors.ENDC+"\n")
	sys.stderr.flush()
	assert False
#        sys.exit(1)

def warn(text):
	sys.stdout.flush()
	sys.stderr.write(bcolors.WARNING+"Warning:"+bcolors.ENDC+" "+text+"\n")
	sys.stderr.flush()
#        sys.exit(1)
    
def nowhitespace(m):
        r = m.replace(" ","")
        r = r.replace("\t","")
        r = r.replace("\r","")
        return r.replace("\n","")

# a useful string of characters
ALPHANUM = "".join([ chr(i)+chr(i).lower() for i in range(ord("A"),ord("Z")+1)])+"0123456789"
    
# returns the maximal subsequence of s that consists only of characters in keep
def keep_char(s,keep):
	return "".join([ r for r in s if r in keep ])

# return true if list a is a prefix of list b
# otherwise return false
def prefix(a,b):
        if len(a)>len(b):
            return False
        for i in range(len(a)):
            if a[i] != b[i]:
                return False
        return True

# copy a list of lists
def unshared_copy(inList):
        if isinstance(inList, list):
            return list( map(unshared_copy, inList) )
        return inList
    
    
# removes all symbols except for alphanumeric ones
def verbs(m):
        r = m
        tospaces = "\t\n\r[];:<>/.(),{}|+=-^#@!"
        for s in tospaces:
            r = r.replace(s," ")
        s = r.replace("  "," ")
        while s != r:
            r = s
            s = r.replace("  "," ")
        r = r.strip(" ")
        return r
    
# returns everything which looks like an ALPHANUMeric function symbol, 
# separated by spaces
def functionsymbols(m):
        r = m+" "
        tospaces="\t\n\r[];:<>)}/.,|+=-^#@!"
        for s in tospaces:
            r = r.replace(s," ")
    
        # remove all names which are not function names, recognized by the 
        # fact that they continue with a space instead of "(" or "{"
        t = ""
        while t != r:
            t = r
            for s in ALPHANUM+"'\"":
                r = r.replace(s+" "," ")
    
        r = r.replace("("," ")
        r = r.replace("{"," ")
        s = r.replace("  "," ")
        while s != r:
            r = s
            s = r.replace("  "," ")
        r = r.strip(" ")
        return r


# returns list of arguments of the top-level function in string s
# arguments in list are strings
def str2args(s):
        if len(s) == 0:
            error("(str2args) empty list")
        l = []
        c = ""
        i = 0
        larg = "("
        rarg = ")"
        sep = ","
        numparen = 0
        while i < len(s):
            if numparen == 0 and s[i] == sep:
                l.append(c)
                c = ""
            else:
                c += s[i]
            
            if s[i] == larg:
                numparen += 1
                i += 1
            elif s[i] == rarg:
                numparen -= 1
                i += 1
    #        elif s[i] == esc:
    #            i += 2
            else:
                i += 1
        l.append(c)
        if numparen != 0:
            error("(str2args) Unbalanced parentheses in string "+s)
    
        return l
    
    
class AnB: 
    def __init__(self):
	    self.rolenames = [] # role names in order of appearance
	    self.T = TermAlgebra() # Create a new Term Algebra T
	    self.prot = Protocol(self.T,"P")

# add a new function symbol to the term algebra T
    def addFunction(self,f,arity):
        self.T.addFunction(f,arity)

    def process_line(self,line):
        if ":" not in line:
            if "knows" in line:
                self.process_line(line.replace("knows",": knows"))
            elif "fresh" in line:
                self.process_line(line.replace("fresh",": fresh"))
            elif "=" in line:
                lhs, rhs = line.split("=")
                lhs = Term(self.T,nowhitespace(lhs))
                rhs = Term(self.T,nowhitespace(rhs))
                vars = lhs.vars()+rhs.vars()
                vars = list(set(vars))
                u = {}
                for v in vars:
                    u[v] = "%"+v
                lhs = lhs.apply(u)
                rhs = rhs.apply(u)
                self.T.addEquation(lhs.to_str()+"="+rhs.to_str())
                # read equations line by line, substitute %variable for every variable
            else:
                error("Invalid input: >>"+line+"<<\n")

        elif nowhitespace(line).lower().startswith("functions:"):
            funcs = line.split(":",1)
            for fs in funcs[1].split(","):
                if nowhitespace(fs) != "":
                    f,arity = fs.split("/",1)
                    f = nowhitespace(f)
                    arity = int(arity)
                    self.addFunction(f,arity)
        elif nowhitespace(line).lower().startswith("equations:"):
            line = line.split(":",1)[1]
            self.process_line(line)
        else:
            AB, m = line.split(":", 1)
            if "->" in AB:
                A,B = AB.split("->", 1)
                A = nowhitespace(A)
                B = nowhitespace(B)
                if A not in self.rolenames:
                    self.rolenames.append(A)
                if B not in self.rolenames:
                    self.rolenames.append(B)
                for rname in self.rolenames:
                    self.prot.add_role(rname,self.rolenames)
    
                self.prot.add_event(B,"receive",{ 'msg': m.strip(" "), 'from': A})
                fsym = functionsymbols(m).split(" ")
                for s in fsym:
                    if s != "" and s not in self.T.Sigma:
                        warn("Unknown function symbol >"+s+"<")
                        self.addFunction(s,-1)
    
                self.prot.add_event(A,"send",{ 'msg': m.strip(" "), 'to': B})
    
            else:
                AB = nowhitespace(AB)
                m = nowhitespace(m)
                if m.startswith("fresh"):
                    stuff = m.split("fresh",1)[1]
                    if stuff[0] == "(" and stuff[-1] == ")":
                        stuff = stuff[1:-1]
                    fsym = functionsymbols(stuff).split(" ")
                    for s in fsym:
                        if s != "":
                            warn("Keyword fresh applied to function symbol >"+s+"<")
                    for i in Term(self.T,"_list("+stuff+")").arguments():
                        if i.is_variable():
                            self.prot.add_event(AB,"fresh",{'msg': nowhitespace(i.to_str())})
                        else:
                            warn("Error fresh keyword applied to a term which is not a variable: "+i.to_str())

                elif m.startswith("knows"):
                    stuff = m.split("knows",1)[1]
                    if stuff[0] == "(" and stuff[-1] == ")":
                        stuff = stuff[1:-1]
                    fsym = functionsymbols(stuff).split(" ")
                    for s in fsym:
                        if s != "" and s not in self.T.Sigma:
                            warn("Unknown function symbol >"+s+"<")
                            self.addFunction(s,-1)

                    knowstuff = []
                    for i in Term(self.T,"_list("+stuff+")").arguments():
                        knowstuff.append(nowhitespace(i.to_str()))
                    self.prot.add_role(AB,knowstuff)
                else:
                    sys.stderr.write("Invalid input: >>"+line+"<<\n")
                    sys.exit(1)
    
    
    def read_from_file(self,arglist):
        filename = ""
        for arg in arglist:
            if arg == "-h" or arg == "--help":
                usage()
                sys.exit(1)
            else: 
                try:
                    f = open(arg)
                    filename = arg 
                except:
                    sys.stderr.write("Unrecognized command line argument.\n")
                    sys.stderr.write("Couldn't open >>"+arg+"<< for reading.\n")
    
        if filename == "":
            filename = "stdin"
            f = sys.stdin

        if filename != "stdin":
            sanitized_filename = nowhitespace(filename).replace("-","_")
            sanitized_filename = sanitized_filename.split("/")[-1]
            sanitized_filename = sanitized_filename.split("\\")[-1]
            if "." in sanitized_filename:
                sanitized_filename = sanitized_filename.split(".")[-2]

            self.prot.protocol_name = keep_char(sanitized_filename,ALPHANUM+"_")
        else:
            self.prot.protocol_name = "noname"

        try:
            for line in f:
                line = line.rstrip('\n') 
                # The following are considered comments
                line = line.split("//",1)[0]
                line = line.split("#",1)[0]
                line = line.split("%",1)[0]
    
                # remove white space
                line = line.replace(" ","")
                line = line.replace("\t","")
                
                # 
                if line != "":
                    self.process_line(line)
    
        finally:
            if type(f) != type([]):
                f.close()
    
    def read_prot(self,arguments):
        self.read_from_file(arguments)

    def output(self):
        self.prot.output()

#### end of class AnB ####

## Knowledge stores terms (of type Term) and uses TermAlgebra to derive new terms from known terms
class Knowledge:
    is_closed = False

    def __init__(self,T):
	    self.KD = {}
	    self.T = T     # TermAlgebra
	    self.DRules = DerivationRules(T) # Contains derivation rules for the subterm convergent equations in T

	    self.M = Maude(self.T)

    def klistup(self,l):
        uni = [ {} ]
        for t in l:
            uniprime = []
            for u in uni:
                uniprime += self.kup(t.apply(u))
            uni = uniprime
        if uni == []:
            return False
        return True
        
           
    def kup(self,t):
        if t.to_str() in self.KD:  # if t has been constructed before, we're done
            return [ {} ]
        elif t.is_ground() and t.is_variable(): # if t is an atomic term (and not in KU, not constructable) then we're doomed. FIX: rename 'var' to 'atomic'
            return [ ]
        elif t.is_variable(): # t is a variable, can be anything
            return [ {} ]
        else:
            # t cannot be obtained by deconstuction, perhaps subpieces of it can: try to construct it from pieces:
            l = []  
            for i in t.arguments(): # t.term['arg'] (but typed as Term)
                r = self.kup(i)
                lres = []
                for u1 in l:
                    for u2 in r:
                        u3 = combinesub(self.T,u1,u2)
                        if u3 != False:
                            lres.append(u3)
            unifiers = []
            for u in lres:
                for c in self.KD:
                    u = self.M.mgusub(c,t.apply(u).to_str())
                    if unifies(u):
                        unifiers.append(u)
            return unifiers
    
    # This will be more useful when annotations need to be added, too
    def add(self,m):
	    if type(m) is str:
		    if m not in self.KD:
			    self.add(Term(self.T,m))
		    return
	    s = m.to_str()
	    if s not in self.KD:
		    self.is_closed = False
		    self.KD[s] = m
    
    # return a list of terms in KD (known terms)
    # the list is a copy of these terms (perhaps [:] is not necessary?)
    def list(self):
	    return self.KD.keys()[:]

    # tells whether a term can be constructed from subterms, purely
    # by applying functions to subterms, no deconstructions!
    # see also: is_derivable above
    def is_constructible(self,t):
        if t.to_str() in self.KD:
            return True  
        elif t.is_variable():
            return False
        else:
        # try to construct it from pieces:
            for i in t.arguments():
                if not self.is_constructible(i):
                    return False
            return True
    
    def minimalKnowledgeList(self):
	    minimal = []
	    for ki in self.KD:
		    t = Term(self.T,ki)
		    if t.is_atomic():
			    minimal.append(ki)
		    else:
			    for kj in t.arguments_as_strlist():
				    if not self.is_constructible(Term(self.T,kj)):
					    minimal.append(ki)
					    break
	    return minimal

    def printKnowledge(self):
        print("  ".join([i for i in self.KD]))
    

### End of class Knowledge

class Maude:
    # T is a TermAlgebra
    def __init__(self,T):
        self.running = False
        successful = False
        # connect to maude
        self.proc = None
        if successful: self.running = True
        self.T = T

    # recall: spaces are significant in maude!!! make spaces.
    def maude_fmod(self,with_equations=False):
        Sigma = self.T.Sigma
        maudefunc = ""
        maudeeq = ""
        for f in Sigma:
            maudefunc += "op "+f+" : "+"A "*Sigma[f]+" -> A .\n"
        maudestr = """fmod MSG is
protecting QID .
sort A .
subsort Qid < A .
"""
        if with_equations: # We don't need equations for syntactic unification 
            vars = []
            for eq in self.T.Equations:
                eq = eq.replace("%","")
                lhs,rhs = eq.split("=")
                lhs = Term(self.T,lhs)
                rhs = Term(self.T,rhs)
                vars = list(set(vars+lhs.vars()+rhs.vars()))
                maudeeq += "eq "+msg2maudestr(lhs.term)+" = "+msg2maudestr(rhs.term)+" .\n"

        maudestr += maudefunc+maudeeq+"\nendfm\n"

        return maudestr


    # reduce a term - could perhaps be useful to rewrite {a}k to enc(a,k)
    def maude_simplify_command(self,term):
        maudestr = self.str2maudestr(term)
        maudestr = "reduce in MSG : "+maudestr+" .\n"
        # maudestr += "quit()\n"  # do this with an explicit maude_quit!
        return maudestr

    def maude_simplify_result(self,output):
    # need to handle the output of maude here
        [ garbage, result ] = output.split("result A:")
        [ result, garbage ] = result.split("\n")
        result = cleanstr(result)
        print("(maude_simplify) Result: "+result)
        return result


    def maude_do_via_file(self,maudestr):
        try:
            f = open("MSG.temp.maude","w")
            f.write(maudestr)
        except:
            error("Preparation for maude interaction failed.")
        finally:
            f.close()

    # Can call maude with "maude file", where file contains all 
    # the queries and then ends with a quit()
    # or can call it with "maude < file", where file contains all 
    # the queries and no quit()
    # The latter will print "Maude> " prompts, the former won't. 
    # The latter might be preferable for intereactive sessions though?
        status, output = getstatusoutput(maude_path+"maude -batch -no-banner -no-ansi-color MSG.temp.maude")
        print("(maude_do) Result: "+output)
        return output

    def maude_do_via_pipe(self,maudecommand):
        if (self.proc == None):
            self.proc = subprocess.Popen(["maude","-no-wrap","-batch","-no-banner","-no-ansi-color"],stdout=subprocess.PIPE,stdin=subprocess.PIPE,stderr=subprocess.STDOUT)
        self.proc.stdin.write(maudecommand)
        dummy = 'red in NAT : 0 .'
        self.proc.stdin.write(dummy+'\n')
        self.proc.stdin.flush()
#        proc.stdin.close()
        s = ""
        while True:
            r = self.proc.stdout.readline()
            s += r
#        s = self.proc.stderr.read()
            if "Warning" in r:
                error("Error in the following command to maude:\n"+maudecommand)
#            elif len(s) > 0:
#                self.proc = None
#                warn("maude - stderr: "+s)
            if 'result Zero: 0' in r:
                break
#        proc.wait() 
        s = s.split(dummy)[0]
        return s

    def maude_do(self,maudecommand):
        return self.maude_do_via_pipe(maudecommand)

    def msg2maudestr(self,msg):
        s = ""
        if msg['type'] == 'var': # and not ground(msg):
            if msg['name'][0] == "%":
                s += msg['name']+":A"
            else:
                s += "'"+msg['name']
        else:
            s = msg['name']
        if msg['type'] == 'func':
            s += "("
            for i in range(msg['arity']):
                if i > 0:
                    s += ","
                s += self.msg2maudestr(msg['arg'][i])
            s += ")"
        return s

    def maudestr2msg(self,s):
        r = s.replace(":A","")
        r = r.replace("'","")
        return r

    def str2maudestr(self,s):
        m = Term(self.T,s)
        return self.msg2maudestr(m.term)

    def maude_unify_command(self,s1,s2):
        m1 = self.str2maudestr(s1)
        m2 = self.str2maudestr(s2)
        # If we have several things to unify simultaneously we can do better than the following. We should then give all those things to maude connected with /\
        command = "unify in MSG : "+m1+" =? "+m2+" ."
        command += "\n"  
        return command

    def maude_unify_result(self,r):
        [ garbage, result ] = r.split("Decision time:")
	# now remove all up to the first newline 
        [ garbage, result ] = result.split("\n",1)
        [ result, garbage ] = result.split("==========================================")

	# finally remove Solution 1 if it occurs:
        if "Solution 1" in result:
            [ garbage, result ] = result.split("Solution 1")

        r = result
        r = r.lstrip("\n")
        return r

    # unifier tool:
    def mgusub(self,s1,s2):
        rho = {}
        maudecommand = self.maude_fmod()+self.maude_unify_command(s1,s2)
    
        r = self.maude_do(maudecommand)
        r = self.maude_unify_result(r)

        if "No unifier" in r:
            return False
        elif "Solution 2" in r:
            error("Unexpected result in mgusub.")
        
        substlist = r.split("\n")

        pre_rho = {}  # contains maude's #1:A, #2:A, .... variables

	# Go through substlist TWICE: First:
	# clean the strings: replace #1:A, ... by the first variable name 
	# that is assigned to them, later remove :A
        for i in range(len(substlist)):
            s = substlist[i]
            if " --> " in s:
                lhs,rhs = s.split(" --> ")
                lhs = nowhitespace(lhs)
                rhs = nowhitespace(rhs)
                if rhs[0] == "#":
                    if rhs not in pre_rho:
                        pre_rho[rhs] = lhs
                    else:
                        rho[self.maudestr2msg(lhs)] = self.maudestr2msg(pre_rho[rhs])
                    # remove this entry from substlist
                    substlist[i] = ""
            elif s == "Maude> ": substlist[i] = ""

        for s in substlist:
            if " --> " in s:
                lhs,rhs = s.split(" --> ")
                lhs = nowhitespace(lhs)
                rhs = nowhitespace(rhs)
                if rhs[0] == "#":
                    error("(mgu-sub) This should not be happening, # have been removed.")
                else:
                    for pr in pre_rho:
                        if pr in rhs:
                            rhs = rhs.replace(pr,pre_rho[pr])
                    if "#" in rhs:
                        warn("Debug info for upcoming error:")
                        print("pre-rho: ",pre_rho)
                        print("rho: ",rho)
                        error("Could not remove # from the following substitutions:\n"+r)
		    # remove :A 
                    lhs = self.maudestr2msg(lhs)
                    rhs = self.maudestr2msg(rhs)
                    if lhs in rho:
                        print("(mgusub) That should not happen: "+lhs+" => "+rhs+", where lhs already in rho, evaluating to "+rho[lhs])
                    rho[lhs] = rhs
            else:
                if s != "":
                    warn("(mgusub) Unexpected line: "+s)
        return rho

##### End of Class Maude

# Substitutes those variables in term by the expression given in the dictionary# assignment that appear as keys in assignment.
def assign(T,term,assignment):
    t = Term(T,term)
    t = t.apply(assignment)
    return t.to_str()

def combinesub(T,rho,sigma):
    if (rho == False) or (sigma == False):
        return False
    # Make a copy of sigma
    u = {}
    for s in sigma:
        u[s] = sigma[s]

    for s in rho:
        if s in u:
            localsubs = mgusub(T,u[s],rho[s])
            if localsubs == False:
                return False
            else:
                u[s] = assign(T,u[s],localsubs)
        else:
            u[s] = rho[s]
    return u

def unifies(u):
    if u == False:
        return False
    else:
        return True


class Term:
    # this class converts strings to the protocol message type:
    # a protocol message is a dictionary: ('type','name','arity','arg'),
    # where type is one of 'func', 'var'
    #       name is the name of the function, or the variable
    #       arity is used for functions: 0 constant, positive, negative: used for lists
    #       list contains further dictionaries, i.e. every list element is a msg

    # T is an instance of the TermAlgebra class
    def __init__(self,T,msgstr = ""):
	    self.T = T
	    if type(msgstr) is str:
		    self.set(msgstr)
	    elif type(msgstr) is dict:
		    self.term = msgstr
	    elif type(msgstr) is Term:
		    print("That's unexpected.")
		    error("Type mismatch type is "+str(type(msgstr)))
	    else:
		    print("Term __init__ received: ",msgstr)
		    error("Type mismatch type is "+str(type(msgstr)))


    def set(self,s):
	    self.term = self._str2msg(s)

    def top_name(self):
	    return self.term['name']

    def is_atomic(self):
	    return (self.is_variable() or len(self.term['arg']) == 0)

    def is_variable(self):
	    if self.term['type'] == 'var':
		    return True
	    return False

    def is_ground(self):
	    return self._ground(self.term)

    def _ground(self,t):
	    if t['type'] == 'func':
		    for i in t['arg']:
			    if not self._ground(i):
				    return False
		    return True
	    else:
		    if t['name'][0] in "%": 
			    return False
		    return True

    # return a list of Term-typed elements. 
    # These elements are the arguments of the top-level function. 
    # They are not copies! 
    # If the term given does not have a top-level function, the empty list is returned.
    def arguments(self):
	    l = []
	    for i in self.term['arg']:
		    l.append(Term(self.T,i))  # This is not a copy
	    return l

    def arguments_as_strlist(self):
	    l = []
	    for i in self.term['arg']:
		    l.append(Term(self.T,i).to_str())  # This is not a copy
	    return l

    # apply a substitution to term. Returns a copy of term with the substitution applied
    # The substituion u is a dictionary mapping strings to strings. The keys of u are names 
    # of variables. The values are the terms (as strings) the variables are assigned to.
    def apply(self,u):
	    m = Term(self.T,self._apply(u,self.term))
	    return m

    def _apply(self,u,t):
        m = {}
        if t['type'] == 'func':
            m['name'] = t['name']
            m['type'] = t['type']
            m['arity'] = t['arity']
            m['arg'] = []
            for i in t['arg']:
                m['arg'].append(self._apply(u,i))
        else:
            if t['name'] in u:
                m = self._str2msg(u[t['name']])
            else:
                m['name'] = t['name']
                m['type'] = t['type']
                m['arity'] = t['arity']
                m['arg'] = unshared_copy(t['arg'])  
        return m
    
    # replaces position pos in self.term by newterm
    # note: newterm is not copied! 
    def replace_in_pos(self,pos,newterm):
        if pos == []:
            self.term = newterm.term
        else:
            m = self.term
            for i in pos[:-1]:
                if m['arity'] >= i:
                    m = m['arg'][i-1]
                else:
                    error("There is no subterm at position "+str(l))
            i = pos[-1]
            if m['arity'] >= i:
                m['arg'][i-1] = newterm.term

    def msg_variable(self,name):
        if name not in self.T.Var:
            self.T.Var[name] = 1
    
        return {'type': 'var', 'name': name,'arity': 0,'arg': []}
    
    def msg_func(self,name,s):
        Sigma = dict(self.T.Sigma)
        Sigma['_list'] = -1 # add a function called _list. To be used internally only
        if name in Sigma:
            l = self.msg_list(s)
            if l['arity'] == Sigma[name] or Sigma[name] == -1:
                return {'type': 'func','name': name, 'arity': l['arity'],'arg': l['arg']}
            else:
                error("Number of arguments ("+str(l['arity'])+") does not match function definition. "+name+" takes "+str(Sigma[name])+" arguments.")
        else:
            error("Unknown function: "+name)
    
    def msg_list(self,s):
        l = str2args(s)
        arg = []
        larg = "("
        rarg = ")"
        for i in l:
            if larg in i:
                name,remainder = i.split(larg,1)
                # remainder should end with rarg
                if remainder[-1] != rarg:
                    error("Syntax error. Expected "+rarg+" as last character in "+remainder)
                arg.append(self.msg_func(name,remainder[:-1]))
            else:
                arg.append(self.msg_variable(i))
    
        return {'type': 'func', 'name': '_list', 'arity': len(arg), 'arg': arg}
    
    # convert a string to a term
    def _str2msg(self,s):
        m = self.msg_list(s)
        if m['arity'] == 1:
            m = self._msgposition(m,[1])
        else: 
            warn("Class Term, _str2msg: Received list of terms instead of a single term")
        return m

    # return the list of variables
    def vars(self):
	    return self.msgvars(self.term)

    # return the list of variables in msg
    def msgvars(self,msg):
        l = []
        if msg['type'] == 'func':
            for i in range(msg['arity']):
                l += self.msgvars(msg['arg'][i])
        elif msg['type'] == 'var':
            l.append(msg['name'])
        return l

    # return the subterm at position l
    # positions are positive integers
    def position(self,l):
	    return self.msgposition(self.term,l)

    def msgposition(self,msg,l):
        return Term(self.T,self._msgposition(msg,l)) ### this is not a copy

    # return the subterm of term msg at position l
    # positions are positive integers
    def _msgposition(self,msg,l):
	    m = msg
	    for i in l:
		    if m['arity'] >= i:
			    m = m['arg'][i-1]
		    else:
			    error("There is no subterm at position "+str(l))
	    return m
    
    def to_str(self):
	    return self.msg2str(self.term)

    def msg2str(self,msg):
        s = msg['name']
        if msg['type'] == 'func':
            s += "("
            for i in range(msg['arity']):
                if i > 0:
                    s += ","
                s += self.msg2str(msg['arg'][i])
            s += ")"
        return s    

    # find all positions of term s in this term
    def find(self,s):
        return self.findstr(self.msg2str(s.term))
    
    # called by find(s).
    # Same as find(), but s here is a string
    def findstr(self,s):
	    return self._msgfindstr(s,self.term,[])

    # called by findstr(s,m), should not be called directly.
    # find all positions of s in msg, where s is a string and msg
    # is a term
    def _msgfindstr(self,s,msg,prefix):
        positions = []
        m = self.msg2str(msg)
        if s == m:
            return [prefix+[]]
        elif msg['type'] == 'var':
            return []
        else:
            for i in range(msg['arity']):
                positions += self._msgfindstr(s,msg['arg'][i],prefix+[i+1])
            return positions
    
    # siblings of the sub-term of this term at position p
    def siblings(self,p):
        siblings = []
        if p == []:
            return []
        parent = p[:-1]
        m = self.position(parent)
        sib = range(1,m.term['arity']+1)
        for i in sib:
            if i != p[-1]:
                siblings.append(parent+[i])
        return siblings
    
#### end of class terms ####


# A term algebra including equations
class TermAlgebra:
    def __init__(self):
        self.Sigma = {} 
        self.Var = {} 
        self.Equations = [] 

    def addFunction(self,f,arity):
        if f not in self.Sigma:
            self.Sigma[f] = arity

    def addEquation(self,eqn):
        self.Equations.append(eqn)
 
#### End of class TermAlgebra
    
# A class used to derive terms from other terms
class DerivationRules:
    def __init__(self,T,Eqns=[]):
	    self.T = T # Term Algebra
	    self.M = Maude(self.T)
	    self.drules = []
	    
	    self.setup_rules()

    def addEquation(self,eqn):
	    self.T.Equations.append(eqn)
	    self.setup_rules()

    def setup_rules(self):
    #### The following are obtained by following steps 1 through 4 in AnBfestschrift
        for arg in self.T.Equations:
            arg = nowhitespace(arg)
            ls,rs = arg.split("=")
            l = Term(self.T,ls)
            r = Term(self.T,rs)
    
            positions = l.find(r)
    
            for p in positions:
                Dp = []
                for i in range(1,len(p)):
                    Dp.append(p[:i])
    
                for pprime in Dp:
                    parent = pprime
                    Cp = []
                    while parent != []:
                        Cp += l.siblings(parent)
                        parent = parent[:-1]
                    
                    r = {}
                    r['lhseq'] = l
#                    r['rhseq'] = r  # not necessary since it is l.position(r['rp'])
                    r['ld'] = l.position(pprime)
                    r['r'] = l.position(p)
                    r['lc'] = []
                    r['rp'] = p # position of r in l
                    r['rdp'] = p[len(pprime):] # relative position of r in ld

                    for q in Cp:
                        r['lc'].append(l.position(q))
                    self.drules.append(r)
 

    # Apply derivation rules to message string m 
    # return a list of results for every applicable rule:
    #  the results contain a list of instantiated terms on the 
    #   left-hand side of the rules (i.e., everything that must be constructed)
    #   the instantiation of the right-hand side of the rule (what can be derived)
    #   and the position of the deconstructed term. 
    def apply(self,m):
        results = []
        # for every deconstruction rule
        for r in self.drules:
            # check to see whether the rule r applies to term m
            u = self.M.mgusub(m,r['ld'].to_str())
            if unifies(u):
                # r applies to m, now apply the same unifier
                # to the KU terms (listed in r['lc']) to apply the rule
                l = []
                for rt in r['lc']:
                    l.append(rt.apply(u))
                # To see whether the result of applying the rule to m
                # is useful (to be decided by the caller), we add the rhs and position to the results list
                rhs = r['r'].apply(u)
                results.append([l,rhs,r['rdp']])
        return results


#### End of class DerivationRules
    

class Event:
    def __init__(self,action,arguments):
        self.action = action
        self.arguments = arguments
        self.next = None

    def append(self,action,arguments):
        ev = Event(action,arguments)
        self.next = ev

    def insert(self,action,arguments):
        ev = Event(action,arguments)
        ev.next = self.next
        self.next = ev

    def _insertBefore(self,action,arguments):
        self.insert(action,arguments)
        tmparg = self.arguments
        tmpaction = self.action
        self.arguments = self.next.arguments
        self.action = self.next.action
        self.next.arguments = tmparg
        self.next.action = tmpaction

    def add_arg(self,key,arg):
        if key not in self.arguments:
            self.arguments[key]=arg

    def update_arg(self,key,arg):
        self.arguments[key]=arg
            

class RoleScript:
    def __init__(self,T,role_name):
        self.role_name = role_name
        self.T = T # TermAlgebra
        self.IK = Knowledge(T) # initial knowledge
        self.firstEvent = Event("create",{}) # list of events
        self.lastEvent = self.firstEvent
        self.currentEvent = self.firstEvent  # keeps track of what event an operation is applied to

# Pretty-printing variables:
        self.arguments_print = { 'msg': "", 'from': " apparently from ", 'to': " intended for ", 'info': " ", 'data': "" }
        self.arguments_order = [ 'msg', 'from', 'to', 'info' ]
        self.arguments_type = {'msg': 'str', 'msglist': 'strlist', 'data': 'omit' }
        self.action_preprint = { 'warning': bcolors.WARNING }
        self.action_print = { 'warning': 'warning:' }
        self.action_postprint = { 'warning': bcolors.ENDC+" " }

    # adds event after last event, keeps current event unchanged
    def add_event(self,action,arguments):
        if self.lastEvent != None:
            self.lastEvent.append(action,arguments)
            self.lastEvent = self.lastEvent.next
        else:
            # we should never be in this case, since roles are generated with a firstEvent called "create"
            warn("Class: RoleScript, add_event: This should not be happening.")
            self.firstEvent = Event(action,arguments)
            self.currentEvent = self.firstEvent
            self.lastEvent = self.firstEvent

    # set currentEvent to the event it points to
    # returns True on success, False when there is no next event
    def next_event(self):
        if self.currentEvent == None:
            return False
        elif self.currentEvent.next == None:
            return False
        else:
            self.currentEvent = self.currentEvent.next
            return True

    def update_action(self,action,key,arg):
        ev = self.find_event_action(action,self.firstEvent)
        if ev == None:
            error("No such action in role script")
        evn = self.find_event_action(action,ev.next)
        if evn != None:
            error("Action is not unique")
        ev.update_arg(key,arg)

    def set_current_event(self,ev):
        self.currentEvent = ev

    # inserts an event after "currentEvent"
    # if currentEvent is None, an error is raised
    # currentEvent points to newly inserted event
    def insert_event(self,action,arguments):
        if self.currentEvent == None:
            error("No current event to insert into.")
        elif self.currentEvent == self.lastEvent:
            self.add_event(action,arguments)
        else:
            self.currentEvent.insert(action,arguments)
        if not self.next_event():
            error("this should be impossible. Error: I1245")

    # inserts an event after "currentEvent"
    # if currentEvent is None, an error is raised
    # difference to above: currentEvent is not moved
    def insertAfter_event(self,action,arguments):
        if self.currentEvent == None:
            error("No current event to insert into.")
        elif self.currentEvent == self.lastEvent:
            self.add_event(action,arguments)
        else:
            self.currentEvent.insert(action,arguments)

    # as above, but inserts an event before "currentEvent"
    # currentEvent is not moved
    def insertBefore_event(self,action,arguments):
        if self.currentEvent == None:
            error("No current event to insert into.")
        elif self.currentEvent == self.firstEvent:
            warn("inserting before the first event!")
            ev = Event(action,arguments)
            ev.next = self.firstEvent
            self.firstEvent = ev
        else:
            self.currentEvent._insertBefore(action,arguments)
            # the following is necessary because _insertBefore copies data
            # and currentEvent ends up pointing to the previous event
            self.currentEvent = self.currentEvent.next

    def find_event_action(self,key,fev):
        ev = fev
        while ev != None:
            if ev.action == key:
                return ev
            ev = ev.next
        return None

    def add_initial_knowledge(self,t):
        self.IK.add(t)

    def set_name(self,name):
        self.role_name = name

    def output(self,onlyspecial=False,noempty=True):
        print("Role "+self.role_name+":")
        print("  Initial Knowledge: ",end='')
        self.IK.printKnowledge()
        print()
        print("  Events:")
        
        ev = self.firstEvent
        while ev != None:
            line = "    "
            if ev.action in self.action_preprint:
                line += self.action_preprint[ev.action]
            if ev.action in self.action_print:
                line += self.action_print[ev.action]
            else:
                line += ev.action
            if ev.action in self.action_postprint:
                line += self.action_postprint[ev.action]
            else:
                line += " "
            for a in self.arguments_order:
                if a in ev.arguments:
                    if a in self.arguments_print:
                        line += self.arguments_print[a]+self.arg2str(ev.arguments,a)
                    else:
                        line += a+": "+str(ev.arguments[a])
            remainder = []
            if not onlyspecial:
                for a in ev.arguments:
                    if a not in self.arguments_order:
                        if (not noempty) or (ev.arguments[a] != ""):
                            if a in self.arguments_print:
                                remainder.append(self.arguments_print[a]+self.arg2str(ev.arguments,a))
                            else:
                                remainder.append(a+": "+self.arg2str(ev.arguments,a))
            if len(remainder) > 0:
                remainder = ",".join(remainder)
                if len(remainder) > 0:
                    remainder = " ("+remainder+")"
            else:
                remainder = ""
            line += remainder
            print(line)
            ev = ev.next

    def arg2str(self,args,a):
        if a in self.arguments_type:
            t = self.arguments_type[a] 
            if t == "str":
                return args[a]
            elif t == "strlist":
                return ", ".join(args[a])
            elif t == "omit":
                return ""
        return str(args[a])

class Protocol:
    def __init__(self,T,name=""):
        self.protocol_name = name
        self.roles = {}
        self.T = T  # TermAlgebra

    def create_role(self,rolename,terms=[]):
        if rolename != "":
            if rolename not in self.roles:
                self.roles[rolename] = RoleScript(self.T,rolename)

            for t in terms:
                self.roles[rolename].add_initial_knowledge(t)
        else:
            error("Empty rolename.")

    def add_role(self,role,terms=[]):
        if type(role) is RoleScript:
            if role.role_name == "":
                A = "Role"+str(len(self.roles)+1)
                role.set_name(A)
            else:
                A = role.role_name
            if A not in self.roles:
                self.roles[A] = role
            else:
                error("Cannot add role name to protocol. Role name exists.")

            for t in terms:
                self.roles[A].add_initial_knowledge(t)

        elif type(role) is str:
            self.create_role(role,terms)

    def add_event(self,rolename,action,arguments):
                self.roles[rolename].add_event(action,arguments)

    def output(self):
        print("Protocol "+self.protocol_name+":")
        for role in self.roles:
            self.roles[role].output()
            print()
        print()


class Checks:
    def __init__(self,T):
        self.T = T    # T is the TermAlgebra
        self.step = 0 # keeps track of the step in which a term is learned or produced
        self.Dterms = {}  # stores positions that a deconstructed or received term appears in

        self.K = Knowledge(T)
        self.Gamma = [] # as in paper. Collects terms that need to be checked


    def annotate(self,m,l):
        s = m.to_str()
        if s in self.Dterms:
            if l not in self.Dterms[s]:
                self.Dterms[s].append(l) 
        else: 
            self.Dterms[s] = [l]
    
    def produce_check(self,m,ann):
        s = self.construct(m,ann)
        if s != False:
            return m.to_str()+"@"+str(ann)+" = "+s
        else:
            return False

    def __call__(self,ABC):
        self.check_protocol(ABC.prot)

    def check_protocol(self,prot):
	    for r in prot.roles:
		    self.__init__(prot.T)
		    self.check_role(prot.roles[r])

## Go through the protocol specification and add checks
    def check_role(self,role):
       #firstWarning = True # this will output the deriviation rules only once
       role.arguments_print['gamma']="^"
       i = role.arguments_order.index('msg')
       role.arguments_order = role.arguments_order[:i+1]+['gamma']+role.arguments_order[i+1:]
       role.arguments_print['eq'] = ""        
       role.arguments_print['current knowledge'] = "" 
       role.arguments_print['annotated knowledge'] = "" 

       role.arguments_type['eq'] = "str"
       role.arguments_type['current knowledge'] = "strlist"
       role.arguments_type['annotated knowledge'] = "strlist"

       role.arguments_order += ['eq','current knowledge','annotated knowledge']
       role.action_print['knows'] = 'role knows:' 

       role.update_action("create",'annotated knowledge',[])
       role.update_action("create",'data',[])

       for m in role.IK.list():
	       self.knows(m)
	       arg = role.firstEvent.arguments['data']
	       arg.append((m,self.step))
	       role.update_action("create",'data',arg)

	       arg = role.firstEvent.arguments['annotated knowledge']
	       arg.append(m+"^"+str(self.step))
	       role.update_action("create",'annotated knowledge',arg)

       role.set_current_event(role.firstEvent)

       step = 0
       while role.next_event(): # this skips the create event
	       step += 1
	       action = role.currentEvent.action
	       arg = role.currentEvent.arguments
	       if action == "receive":
		       m = arg['msg']
		       list_of_checks = self.learn(m)
		       role.currentEvent.add_arg("gamma",self.step)
		       role.insert_event("knows",{'current knowledge': self.K.minimalKnowledgeList()})
		       for check_term in list_of_checks:
			       role.insert_event("check",{'eq': check_term})
	       elif action == "fresh":
		       m = arg['msg']
		       self.knows(m)
		       role.currentEvent.add_arg("gamma",self.step)
	       elif action == "send":  
		       m = Term(self.T,arg['msg'])
		       if not self.K.is_constructible(m):
			       role.insertBefore_event("warning",{'msg':m.to_str(),'info':"not derivable",'knowledge':self.K.list()})
	       else:
		       if 'msg' in arg:
			       m = arg['msg']
		       else: 
			       m = str(arg)
		       role.insertBefore_event("warning",{'info':"unknown action '"+action+"' with argument "+m})

       if len(self.Gamma) > 0:
	       role.arguments_print['msglist']=""
	       role.arguments_order.append('msglist')
	       role.action_print['not checked'] = 'role cannot check:' 
	       role.add_event("not checked",{'msglist': self.listMinimalGamma()})


    def closeKDG(self):
	    while not self.K.is_closed:
		    self.K.is_closed = True
		    for m in self.K.KD.keys():
			    t = self.K.KD[m]
			    if not t.is_variable():  # we cannot deconstruct vars
				    # Apply all derivation rules to m
				    for l,rhs,rdp in self.K.DRules.apply(m):
					    rhspositions = []
					    for i in self.Dterms[m]:
						    rhspositions.append(i+rdp)
					    # need to find all positions rhs.to_str() could have 
					    # and test whether one of them is missing
					    if rhs.to_str() not in self.K.KD:
						    if self.K.klistup(l):  
							    self.K.add(rhs) # sets K.is_closed to False
							    for i in rhspositions:
								    self.annotate(rhs,i)
								    self.Gamma.append([rhs,i])
					    else:
						    for i in rhspositions:
							    if i not in self.Dterms[rhs.to_str()]:
								    self.annotate(rhs,i)
								    self.Gamma.append([rhs,i])

    
    def construct(self,t,ann):
	    s = t.to_str()
	    if t.is_variable():
		    if s in self.K.KD:
			    for i in self.Dterms[s]:
				    if not prefix(ann,i):
					    return s+"^"+str(i)+""
				    return False
		    return False
	    else:
		    # try to construct it from pieces:
		    if s in self.K.KD:
			    for i in self.Dterms[s]:
				    if not prefix(ann,i):
					    return s+"^"+str(i)+""
		    r = t.top_name()+"("
		    for i in t.arguments():
			    res = self.construct(i,ann)
			    if res == False:
				    return False
			    if r[-1] != "(": # check whether this is the first argument to be added
				    r += ","+res # it is not, thus put a comma to separate from prev arg
			    else:
				    r += res # it's the first arg, no preceding comma necessary
		    r += ")"
		    return r
    
    # we return the smallest elements of Gamma which have not been checked
    def minimalGamma(self):
	    minimal = []
	    for gi in self.Gamma:
		    noprefix = True
		    for gj in self.Gamma:
			    if gi != gj and prefix(gi[1],gj[1]): # note the order
				    noprefix = False
				    break
		    if noprefix: minimal.append(gi)
	    return minimal

    def listMinimalGamma(self):
	    return [ i[0].to_str()+"@"+str(i[1]) for i in self.minimalGamma() ]

    def learn(self,s):
	    list_of_checks = []
	    self.knows(s,True) # True indicates that it has been learned, causes Gamma.append(...)
	    Gammaprime = []
	    for ghost in self.Gamma:
		    check = self.produce_check(ghost[0],ghost[1])
		    if check == False:
			    Gammaprime.append(ghost)
		    else:
			    list_of_checks.append(check)
	    self.Gamma = Gammaprime
	    return list_of_checks

    def knows(self,s,learned=False):
	    self.step += 1
	    m = Term(self.T,s)
	    self.K.add(m)
	    self.annotate(m,[self.step])
	    if learned: self.Gamma.append([m,[self.step]])                
	    self.closeKDG()


####### MAIN ##########
if __name__ == "__main__":
    
   arguments = sys.argv[1:]
   ABC = AnB()
   ABC.read_prot(arguments)
   check = Checks(ABC.T)
   check(ABC)
   ABC.output()

